import fnmatch
import logging
import mimetypes
import tempfile
import urllib
from os import listdir
from os.path import join

from kodexa.model import Document, DocumentMetadata


class FolderConnector:

    @staticmethod
    def get_name():
        return "folder"

    def __init__(self, path, file_filter="*"):
        self.path = path
        self.file_filter = file_filter

        if not self.path:
            raise ValueError('You must provide a path')

        self.files = self.__get_files__()
        self.index = 0

    def get_source(self, document):
        return open(join(self.path, document.metadata.source_path), 'rb')

    def __iter__(self):
        return self

    def __next__(self):
        if self.index > len(self.files) - 1:
            raise StopIteration
        else:
            self.index += 1
            return Document(DocumentMetadata(
                {"source_path": self.files[self.index - 1], "connector": self.get_name(),
                 "mime_type": mimetypes.guess_type(self.files[self.index - 1]),
                 "connector_options": {"path": self.path, "file_filter": self.file_filter}}))

    def __get_files__(self):
        return [f for f in listdir(self.path) if
                fnmatch.fnmatch(f, self.file_filter)]


class FileHandleConnector:

    @staticmethod
    def get_name():
        return "file-handle"

    def __init__(self, file):
        self.file = file
        self.index = 0
        self.completed = False

    def get_source(self, document):
        return open(join(self.file, document.metadata.source_path), 'rb')

    def __iter__(self):
        return self

    def __next__(self):
        if self.completed:
            raise StopIteration
        else:
            return Document(DocumentMetadata(
                {"source_path": self.file, "connector": self.get_name(),
                 "mime_type": mimetypes.guess_type(self.file),
                 "connector_options": {"file": self.file}}))


class UrlConnector:

    @staticmethod
    def get_name():
        return "url"

    def __init__(self, url, headers={}):
        self.url = url
        self.headers = headers
        self.index = 0
        self.completed = False

    def get_source(self, document):
        if 'headers' in document.metadata.connector_options:
            opener = urllib.request.build_opener()
            for header in document.metadata.connector_options['headers']:
                opener.addheaders = [(header, document.metadata.connector_options['headers'][header])]
            urllib.request.install_opener(opener)
        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            urllib.request.urlretrieve(document.metadata.connector_options['url'], tmp_file.name)

            return open(tmp_file.name, 'rb')

    def __iter__(self):
        return self

    def __next__(self):
        if self.completed:
            raise StopIteration
        else:
            self.completed = True
            return Document(DocumentMetadata(
                {"connector": self.get_name(),
                 "connector_options": {"url": self.url, "headers": self.headers}}))


registered_connectors = {}


def get_connectors():
    return registered_connectors.keys()


def get_connector(connector, options):
    if connector in registered_connectors:
        logging.info("Getting registered connector")
        return registered_connectors[connector](**options)
    else:
        logging.info(f"Unable to find connector {connector}")
        return None


def add_connector(connector):
    registered_connectors[connector.get_name()] = connector


def get_source(document):
    connector = get_connector(document.metadata.connector,
                              document.metadata.connector_options if "connector_options" in document.metadata else {})
    return connector.get_source(document)


add_connector(FolderConnector)
add_connector(FileHandleConnector)
add_connector(UrlConnector)



class InMemoryDocumentSink:
    """
    An in-memory document sink can be used for testing where you want to capture a set of the documents as basic
    list in-memory and then access them
    """

    def __init__(self):
        self.documents = []

    @staticmethod
    def get_name():
        return "In-Memory Document"

    def sink(self, document):
        """
        Adds the document to the sink

        :param document: document to add
        """
        self.documents.append(document)

    def get_document(self, index):
        """
        Get document at given index

        :param index: index to get the document at
        """
        return self.documents[index]



# generated by datamodel-codegen:
#   filename:  kodexa-swagger.json
#   timestamp: 2020-04-19T15:10:00+00:00

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class Model(BaseModel):
    __root__: 'Any'


class BotEventTrigger(BaseModel):
    cellId: Optional[str] = None
    event: Optional[str] = None
    type: Optional[str] = None


class CellProcessor(BaseModel):
    condition: Optional[str] = None
    configuration: Optional[Dict[str, Any]] = None
    targetCellId: Optional[str] = None
    type: Optional[str] = None


class SessionState(Enum):
    OPEN = 'OPEN'
    CLOSED = 'CLOSED'


class SessionState1(Enum):
    OPEN = 'OPEN'
    CLOSED = 'CLOSED'


class Type(Enum):
    START_EXECUTION = 'START_EXECUTION'
    STEP_UPDATE = 'STEP_UPDATE'
    NEW_SESSION = 'NEW_SESSION'
    CLOSE_SESSION = 'CLOSE_SESSION'
    BOT_NOTIFICATION = 'BOT_NOTIFICATION'


class Status(Enum):
    PENDING = 'PENDING'
    RUNNING = 'RUNNING'
    SUCCEEDED = 'SUCCEEDED'
    FAILED = 'FAILED'
    REQUESTED = 'REQUESTED'


class Status1(Enum):
    PENDING = 'PENDING'
    RUNNING = 'RUNNING'
    SUCCEEDED = 'SUCCEEDED'
    FAILED = 'FAILED'
    REQUESTED = 'REQUESTED'


class SessionState2(Enum):
    OPEN = 'OPEN'
    CLOSED = 'CLOSED'


class SessionState3(Enum):
    OPEN = 'OPEN'
    CLOSED = 'CLOSED'


class CloudStore(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None


class ContentType(Enum):
    DOCUMENT = 'DOCUMENT'
    NATIVE = 'NATIVE'


class ContentObject(BaseModel):
    contentType: Optional['ContentType'] = None
    id: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    name: Optional[str] = None
    tags: Optional[List[str]] = None


class ContentServiceReference(BaseModel):
    attachSource: Optional[bool] = None
    contentServiceId: Optional[str] = None
    contentServiceVersionId: Optional[str] = None
    ref: Optional[str] = None


class DeploymentMetadata(BaseModel):
    accessToken: Optional[str] = None
    botId: Optional[str] = None
    type: str


class EmailDeployment(DeploymentMetadata):
    accessToken: Optional[str] = None
    address: Optional[str] = None
    botId: Optional[str] = None
    type: str


class EmailLookup(BaseModel):
    email: Optional[str] = None
    id: Optional[str] = None


class ExceptionDetails(BaseModel):
    errorMessage: Optional[str] = None
    errorType: Optional[str] = None
    executedVersion: Optional[str] = None
    help: Optional[str] = None
    message: Optional[str] = None
    stackTrace: Optional[List[Dict[str, Any]]] = None
    statusCode: Optional[int] = None


class ExceptionReport(BaseModel):
    description: Optional[str] = None
    exceptionUuid: Optional[str] = None
    message: Optional[str] = None
    platform: Optional[str] = None


class Feedback(BaseModel):
    createdBy: Optional[str] = None
    createdOn: Optional[str] = None
    feedback: Optional[str] = None
    id: Optional[str] = None
    updatedBy: Optional[str] = None
    updatedOn: Optional[str] = None
    uuid: Optional[str] = None


class File(BaseModel):
    absolute: Optional[bool] = None
    absoluteFile: Optional['File'] = None
    absolutePath: Optional[str] = None
    canonicalFile: Optional['File'] = None
    canonicalPath: Optional[str] = None
    directory: Optional[bool] = None
    file: Optional[bool] = None
    freeSpace: Optional[int] = None
    hidden: Optional[bool] = None
    name: Optional[str] = None
    parent: Optional[str] = None
    parentFile: Optional['File'] = None
    path: Optional[str] = None
    totalSpace: Optional[int] = None
    usableSpace: Optional[int] = None


class Type1(Enum):
    PIPELINE = 'PIPELINE'
    ACTION = 'ACTION'
    DECISION = 'DECISION'


class InputStream(BaseModel):
    pass


class LoginRequest(BaseModel):
    email: Optional[str] = None
    password: Optional[str] = None


class Role(Enum):
    OWNER = 'OWNER'
    READ = 'READ'
    WRITE = 'WRITE'


class Organization(BaseModel):
    createdBy: Optional[str] = None
    createdOn: Optional[str] = None
    description: Optional[str] = None
    id: Optional[str] = None
    name: Optional[str] = None
    publicAccess: Optional[bool] = None
    slug: Optional[str] = None
    updatedBy: Optional[str] = None
    updatedOn: Optional[str] = None
    uuid: Optional[str] = None


class OrganizationDashboard(BaseModel):
    pipelineCount: Optional[int] = None
    requestCount: Optional[int] = None
    serviceCount: Optional[int] = None
    userCount: Optional[int] = None


class PasswordChange(BaseModel):
    newPassword: Optional[str] = None
    oldPassword: Optional[str] = None


class PipelineStepMetadata(BaseModel):
    name: Optional[str] = None
    options: Optional[Dict[str, Any]] = None
    organizationSlug: Optional[str] = None
    slug: Optional[str] = None


class PlatformOverview(BaseModel):
    buildTime: Optional[str] = None
    commitId: Optional[str] = None
    dsn: Optional[str] = None
    environment: Optional[str] = None
    hostName: Optional[str] = None
    name: Optional[str] = None
    release: Optional[str] = None
    version: Optional[str] = None


class RegisterUser(BaseModel):
    email: Optional[str] = None
    firstName: Optional[str] = None
    lastName: Optional[str] = None


class SlackDeployment(DeploymentMetadata):
    accessToken: Optional[str] = None
    botId: Optional[str] = None
    channelId: Optional[str] = None
    type: str


class Sort(BaseModel):
    empty: Optional[bool] = None
    sorted: Optional[bool] = None
    unsorted: Optional[bool] = None


class SubscriptionResponse(BaseModel):
    alert: Optional[str] = None
    message: Optional[str] = None


class Synapse(BaseModel):
    configuration: Optional[Dict[str, Any]] = None
    targetCellIds: Optional[List[str]] = None
    type: Optional[str] = None


class Timestamp(BaseModel):
    date: Optional[int] = None
    day: Optional[int] = None
    hours: Optional[int] = None
    minutes: Optional[int] = None
    month: Optional[int] = None
    nanos: Optional[int] = None
    seconds: Optional[int] = None
    time: Optional[int] = None
    timezoneOffset: Optional[int] = None
    year: Optional[int] = None


class URI(BaseModel):
    absolute: Optional[bool] = None
    authority: Optional[str] = None
    fragment: Optional[str] = None
    host: Optional[str] = None
    opaque: Optional[bool] = None
    path: Optional[str] = None
    port: Optional[int] = None
    query: Optional[str] = None
    rawAuthority: Optional[str] = None
    rawFragment: Optional[str] = None
    rawPath: Optional[str] = None
    rawQuery: Optional[str] = None
    rawSchemeSpecificPart: Optional[str] = None
    rawUserInfo: Optional[str] = None
    scheme: Optional[str] = None
    schemeSpecificPart: Optional[str] = None
    userInfo: Optional[str] = None


class URL(BaseModel):
    authority: Optional[str] = None
    content: Optional[Dict[str, Any]] = None
    defaultPort: Optional[int] = None
    file: Optional[str] = None
    host: Optional[str] = None
    path: Optional[str] = None
    port: Optional[int] = None
    protocol: Optional[str] = None
    query: Optional[str] = None
    ref: Optional[str] = None
    userInfo: Optional[str] = None


class UserActivation(BaseModel):
    firstName: Optional[str] = None
    lastName: Optional[str] = None
    password: Optional[str] = None


class View(BaseModel):
    contentType: Optional[str] = None


class WebHookDeployment(DeploymentMetadata):
    accessToken: Optional[str] = None
    botId: Optional[str] = None
    type: str
    uuid: Optional[str] = None


class AccessToken(BaseModel):
    createdBy: Optional[str] = None
    createdOn: Optional[str] = None
    id: Optional[str] = None
    name: Optional[str] = None
    organization: Optional['Organization'] = None
    token: Optional[str] = None
    updatedBy: Optional[str] = None
    updatedOn: Optional[str] = None
    uuid: Optional[str] = None


class ByteArrayResource(BaseModel):
    byteArray: Optional[str] = None
    description: Optional[str] = None
    file: Optional['File'] = None
    filename: Optional[str] = None
    inputStream: Optional['InputStream'] = None
    open: Optional[bool] = None
    readable: Optional[bool] = None
    uri: Optional['URI'] = None
    url: Optional['URL'] = None


class CloudExecutionStep(BaseModel):
    contentObjects: Optional[List['ContentObject']] = None
    contentServiceReference: Optional['ContentServiceReference'] = None
    context: Optional[Dict[str, Any]] = None
    end: Optional[datetime] = None
    exceptionDetails: Optional['ExceptionDetails'] = None
    id: Optional[str] = None
    inputId: Optional[str] = None
    options: Optional[Dict[str, Any]] = None
    outputId: Optional[str] = None
    processingTime: Optional[int] = None
    start: Optional[datetime] = None
    status: Optional['Status1'] = None
    stores: Optional[List['CloudStore']] = None


class CloudPipeline(BaseModel):
    exceptions: Optional[List['ExceptionDetails']] = None
    id: Optional[str] = None
    ref: Optional[str] = None
    steps: Optional[List['CloudExecutionStep']] = None
    valid: Optional[bool] = None


class CloudSession(BaseModel):
    contentObjects: Optional[List['ContentObject']] = None
    id: str
    lastAccessed: Optional[int] = None
    pipelineTemplates: Optional[List['CloudPipeline']] = None
    sessionState: 'SessionState3'
    token: Optional[str] = None
    type: str


class ContentService(BaseModel):
    createdBy: Optional[str] = None
    createdOn: Optional[str] = None
    description: Optional[str] = None
    id: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    name: Optional[str] = None
    organization: Optional['Organization'] = None
    publicAccess: Optional[bool] = None
    ref: Optional[str] = None
    slug: Optional[str] = None
    updatedBy: Optional[str] = None
    updatedOn: Optional[str] = None
    uuid: Optional[str] = None


class ContentServiceVersion(BaseModel):
    contentService: Optional['ContentService'] = None
    createdBy: Optional[str] = None
    createdOn: Optional[str] = None
    id: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    position: Optional[int] = None
    publicAccess: Optional[bool] = None
    updatedBy: Optional[str] = None
    updatedOn: Optional[str] = None
    uuid: Optional[str] = None
    version: Optional[str] = None


class Deployment(BaseModel):
    createdBy: Optional[str] = None
    createdOn: Optional[str] = None
    description: Optional[str] = None
    id: Optional[str] = None
    metadata: Optional['DeploymentMetadata'] = None
    name: Optional[str] = None
    organization: Optional['Organization'] = None
    updatedBy: Optional[str] = None
    updatedOn: Optional[str] = None
    uuid: Optional[str] = None


class GraphCell(BaseModel):
    id: Optional[str] = None
    options: Optional[Dict[str, Any]] = None
    post: Optional[List['CellProcessor']] = None
    pre: Optional[List['CellProcessor']] = None
    ref: Optional[str] = None
    synapses: Optional[List['Synapse']] = None
    type: Optional['Type1'] = None


class Pageable(BaseModel):
    offset: Optional[int] = None
    pageNumber: Optional[int] = None
    pageSize: Optional[int] = None
    paged: Optional[bool] = None
    sort: Optional['Sort'] = None
    unpaged: Optional[bool] = None


class PipelineMetadata(BaseModel):
    steps: Optional[List['PipelineStepMetadata']] = None


class PlatformUser(BaseModel):
    activated: Optional[bool] = None
    createdBy: Optional[str] = None
    createdOn: Optional[str] = None
    email: Optional[str] = None
    firstName: Optional[str] = None
    id: Optional[str] = None
    lastName: Optional[str] = None
    passwordResetDate: Optional['Timestamp'] = None
    platformAdmin: Optional[bool] = None
    updatedBy: Optional[str] = None
    updatedOn: Optional[str] = None
    uuid: Optional[str] = None


class BotMetadata(BaseModel):
    createSession: Optional[bool] = None
    graphCells: Optional[List['GraphCell']] = None
    triggers: Optional[List['BotEventTrigger']] = None


class CloudContentServiceSession(CloudSession):
    contentObjects: Optional[List['ContentObject']] = None
    contentService: Optional['ContentService'] = None
    id: str
    lastAccessed: Optional[int] = None
    pipelineTemplates: Optional[List['CloudPipeline']] = None
    sessionState: 'SessionState1'
    token: Optional[str] = None
    type: str


class CloudEvent(BaseModel):
    contentObjects: Optional[List['ContentObject']] = None
    executionId: Optional[str] = None
    id: str
    inputId: Optional[str] = None
    payload: Optional[Dict[str, Any]] = None
    sessionId: str
    source: Optional[Dict[str, Any]] = None
    step: Optional['CloudExecutionStep'] = None
    subType: Optional[str] = None
    type: 'Type'


class CloudExecution(BaseModel):
    contentObjects: Optional[List['ContentObject']] = None
    context: Optional[Dict[str, Any]] = None
    customOptions: Optional[Dict[str, Any]] = None
    end: Optional[datetime] = None
    exceptionDetails: Optional['ExceptionDetails'] = None
    id: Optional[str] = None
    inputId: Optional[str] = None
    numberOfSteps: Optional[int] = None
    outputId: Optional[str] = None
    pipelineTemplateId: Optional[str] = None
    processingTime: Optional[int] = None
    sessionId: Optional[str] = None
    start: Optional[datetime] = None
    status: Optional['Status'] = None
    steps: Optional[List['CloudExecutionStep']] = None
    stepsCompleted: Optional[int] = None
    stores: Optional[List['CloudStore']] = None


class Membership(BaseModel):
    createdBy: Optional[str] = None
    createdOn: Optional[str] = None
    id: Optional[str] = None
    organization: Optional['Organization'] = None
    role: Optional['Role'] = None
    updatedBy: Optional[str] = None
    updatedOn: Optional[str] = None
    user: Optional['PlatformUser'] = None
    uuid: Optional[str] = None


class PageOfAccessToken(BaseModel):
    content: Optional[List['AccessToken']] = None
    empty: Optional[bool] = None
    first: Optional[bool] = None
    last: Optional[bool] = None
    number: Optional[int] = None
    numberOfElements: Optional[int] = None
    pageable: Optional['Pageable'] = None
    size: Optional[int] = None
    sort: Optional['Sort'] = None
    totalElements: Optional[int] = None
    totalPages: Optional[int] = None


class PageOfCloudEvent(BaseModel):
    content: Optional[List['CloudEvent']] = None
    empty: Optional[bool] = None
    first: Optional[bool] = None
    last: Optional[bool] = None
    number: Optional[int] = None
    numberOfElements: Optional[int] = None
    pageable: Optional['Pageable'] = None
    size: Optional[int] = None
    sort: Optional['Sort'] = None
    totalElements: Optional[int] = None
    totalPages: Optional[int] = None


class PageOfContentService(BaseModel):
    content: Optional[List['ContentService']] = None
    empty: Optional[bool] = None
    first: Optional[bool] = None
    last: Optional[bool] = None
    number: Optional[int] = None
    numberOfElements: Optional[int] = None
    pageable: Optional['Pageable'] = None
    size: Optional[int] = None
    sort: Optional['Sort'] = None
    totalElements: Optional[int] = None
    totalPages: Optional[int] = None


class PageOfContentServiceVersion(BaseModel):
    content: Optional[List['ContentServiceVersion']] = None
    empty: Optional[bool] = None
    first: Optional[bool] = None
    last: Optional[bool] = None
    number: Optional[int] = None
    numberOfElements: Optional[int] = None
    pageable: Optional['Pageable'] = None
    size: Optional[int] = None
    sort: Optional['Sort'] = None
    totalElements: Optional[int] = None
    totalPages: Optional[int] = None


class PageOfDeployment(BaseModel):
    content: Optional[List['Deployment']] = None
    empty: Optional[bool] = None
    first: Optional[bool] = None
    last: Optional[bool] = None
    number: Optional[int] = None
    numberOfElements: Optional[int] = None
    pageable: Optional['Pageable'] = None
    size: Optional[int] = None
    sort: Optional['Sort'] = None
    totalElements: Optional[int] = None
    totalPages: Optional[int] = None


class PageOfMembership(BaseModel):
    content: Optional[List['Membership']] = None
    empty: Optional[bool] = None
    first: Optional[bool] = None
    last: Optional[bool] = None
    number: Optional[int] = None
    numberOfElements: Optional[int] = None
    pageable: Optional['Pageable'] = None
    size: Optional[int] = None
    sort: Optional['Sort'] = None
    totalElements: Optional[int] = None
    totalPages: Optional[int] = None


class PageOfOrganization(BaseModel):
    content: Optional[List['Organization']] = None
    empty: Optional[bool] = None
    first: Optional[bool] = None
    last: Optional[bool] = None
    number: Optional[int] = None
    numberOfElements: Optional[int] = None
    pageable: Optional['Pageable'] = None
    size: Optional[int] = None
    sort: Optional['Sort'] = None
    totalElements: Optional[int] = None
    totalPages: Optional[int] = None


class PageOfPlatformUser(BaseModel):
    content: Optional[List['PlatformUser']] = None
    empty: Optional[bool] = None
    first: Optional[bool] = None
    last: Optional[bool] = None
    number: Optional[int] = None
    numberOfElements: Optional[int] = None
    pageable: Optional['Pageable'] = None
    size: Optional[int] = None
    sort: Optional['Sort'] = None
    totalElements: Optional[int] = None
    totalPages: Optional[int] = None


class Pipeline(BaseModel):
    createdBy: Optional[str] = None
    createdOn: Optional[str] = None
    description: Optional[str] = None
    id: Optional[str] = None
    metadata: Optional['PipelineMetadata'] = None
    name: Optional[str] = None
    organization: Optional['Organization'] = None
    publicAccess: Optional[bool] = None
    ref: Optional[str] = None
    slug: Optional[str] = None
    updatedBy: Optional[str] = None
    updatedOn: Optional[str] = None
    uuid: Optional[str] = None


class ServiceRequest(BaseModel):
    contentService: Optional['ContentService'] = None
    contentServiceVersion: Optional['ContentServiceVersion'] = None
    createdBy: Optional[str] = None
    createdOn: Optional[str] = None
    elapsedTime: Optional[int] = None
    id: Optional[str] = None
    internalRequestId: Optional[str] = None
    organization: Optional['Organization'] = None
    pipeline: Optional['Pipeline'] = None
    updatedBy: Optional[str] = None
    updatedOn: Optional[str] = None
    uuid: Optional[str] = None


class Bot(BaseModel):
    createdBy: Optional[str] = None
    createdOn: Optional[str] = None
    description: Optional[str] = None
    id: Optional[str] = None
    metadata: Optional['BotMetadata'] = None
    name: Optional[str] = None
    organization: Optional['Organization'] = None
    updatedBy: Optional[str] = None
    updatedOn: Optional[str] = None
    uuid: Optional[str] = None


class CloudBotSession(CloudSession):
    bot: Optional['Bot'] = None
    botState: Optional[Dict[str, Any]] = None
    contentObjects: Optional[List['ContentObject']] = None
    id: str
    lastAccessed: Optional[int] = None
    pipelineTemplates: Optional[List['CloudPipeline']] = None
    sessionState: 'SessionState'
    token: Optional[str] = None
    type: str


class CloudPipelineSession(CloudSession):
    contentObjects: Optional[List['ContentObject']] = None
    id: str
    lastAccessed: Optional[int] = None
    pipeline: Optional['Pipeline'] = None
    pipelineTemplates: Optional[List['CloudPipeline']] = None
    sessionState: 'SessionState2'
    token: Optional[str] = None
    type: str


class PageOfBot(BaseModel):
    content: Optional[List['Bot']] = None
    empty: Optional[bool] = None
    first: Optional[bool] = None
    last: Optional[bool] = None
    number: Optional[int] = None
    numberOfElements: Optional[int] = None
    pageable: Optional['Pageable'] = None
    size: Optional[int] = None
    sort: Optional['Sort'] = None
    totalElements: Optional[int] = None
    totalPages: Optional[int] = None


class PageOfPipeline(BaseModel):
    content: Optional[List['Pipeline']] = None
    empty: Optional[bool] = None
    first: Optional[bool] = None
    last: Optional[bool] = None
    number: Optional[int] = None
    numberOfElements: Optional[int] = None
    pageable: Optional['Pageable'] = None
    size: Optional[int] = None
    sort: Optional['Sort'] = None
    totalElements: Optional[int] = None
    totalPages: Optional[int] = None


class PageOfServiceRequest(BaseModel):
    content: Optional[List['ServiceRequest']] = None
    empty: Optional[bool] = None
    first: Optional[bool] = None
    last: Optional[bool] = None
    number: Optional[int] = None
    numberOfElements: Optional[int] = None
    pageable: Optional['Pageable'] = None
    size: Optional[int] = None
    sort: Optional['Sort'] = None
    totalElements: Optional[int] = None
    totalPages: Optional[int] = None


File.update_forward_refs()



import logging
import sys
import traceback
import uuid
from inspect import signature
from io import StringIO
from typing import List
from uuid import uuid4

from kodexa import Document
from kodexa_cloud import ContentObject


def new_id():
    return str(uuid.uuid4())


class InMemoryContentProvider:
    """
    A content provider is used to support getting content (documents or native) to
    and from the pipeline
    """

    def __init__(self):
        self.content_objects = {}

    def get_content(self, content_object: ContentObject):
        return self.content_objects[content_object.id]

    def put_content(self, content_object: ContentObject, content):
        self.content_objects[content_object.id] = content


class PipelineContext:
    """
    Pipeline context is created when you create a pipeline and it provides a way to access information about the
    pipeline that is running.  It can be made available to steps/functions so they can interact with it.

    It also provides access to the 'stores' that have been added to the pipeline
    """

    def __init__(self, content_provider=InMemoryContentProvider(), existing_content_objects=None,
                 context=None):
        if context is None:
            context = {}
        if existing_content_objects is None:
            existing_content_objects = []
        self.transaction_id = str(uuid4())
        self.stores = {}
        self.statistics = PipelineStatistics()
        self.output_document = None
        self.content_objects: List[ContentObject] = existing_content_objects
        self.content_provider = content_provider
        self.context = context

    def get_context(self):
        return self.context

    def get_content_objects(self) -> List[ContentObject]:
        return self.content_objects

    def get_content(self, content_object: ContentObject):
        self.content_provider.get_content(content_object)

    def put_content(self, content_object: ContentObject, content):
        self.content_provider.put_content(content_object, content)

    def add_store(self, name, store):
        """
        Add a store with given name to the context

        :param name: the name to refer to the store with
        :param store: the instance of the store
        """
        self.stores[name] = store

    def get_store_names(self):
        """
        Return the list of store names in context

        :return: the list of store names
        """
        return list(self.stores.keys())

    def set_output_document(self, output_document):
        """
        Set the output document from the pipeline

        :param output_document: the final output document from the pipeline
        :return: the final output document
        """
        self.output_document = output_document

    def get_store(self, name, default=None):
        """
        Get a store with given name from the context

        :param name: the name to refer to the store with
        :param default: optionally the default to create the store as if it isn't there
        :return: the store, or None is not available
        """
        store = self.stores[name] if name in self.stores else None

        if not store and default:
            self.stores[name] = default

        return self.stores[name]


class Pipeline:
    """
    A pipeline represents a way to bring together parts of the kodexa framework to solve a specific problem.

    When you create a Pipeline you must provide the connector that will be used to source the documents.

        >>> pipeline = Pipeline(FolderConnector(path='/tmp/', file_filter='example.pdf'))

    :param connector: the connector that will be the starting point for the pipeline
    :param name: the name of the pipeline (default 'Default')
    :param stop_on_exception: Should the pipeline raise exceptions and stop (default True)
    :param logging_level: The logging level of the pipeline (default INFO)
    """

    def __init__(self, connector, name="Default", stop_on_exception=True, logging_level=logging.INFO):
        logging.info(f"Initializing a new pipeline {name}")

        if isinstance(connector, Document):
            self.connector = [connector]
        else:
            self.connector = connector

        self.steps = []
        self.sink = None
        self.name = name
        self.context = PipelineContext()
        self.stop_on_exception = stop_on_exception
        self.logging_level = logging_level

    def add_store(self, name, store):
        """
        Add the store to the pipeline so that it is available to the pipeline

            >>> pipeline = Pipeline(FolderConnector(path='/tmp/', file_filter='example.pdf'))
            >>> pipeline.add_store("test-store", InMemoryObjectStore())

        :param name: the name of the store (to refer to it)
        :param store: the store that should be added
        """
        self.context.add_store(name, store)

        return self

    def add_step(self, step):
        """
        Add the given step to the current pipeline

            >>> pipeline = Pipeline(FolderConnector(path='/tmp/', file_filter='example.pdf'))
            >>> pipeline.add_step(ExampleStep())

        Note that it is also possible to add a function as a step, for example

            >>> def my_function(doc):
            >>>      doc.metadata.fishstick = 'foo'
            >>>      return doc
            >>> pipeline.add_step(my_function)

        :param step: the step to add
        """
        if callable(step):
            logging.info(f"Adding new step function {step.__name__} to pipeline {self.name}")
        else:
            logging.info(f"Adding new step {step.get_name()} to pipeline {self.name}")

        self.steps.append(step)

        return self

    def set_sink(self, sink):
        """
        Set the sink you wish to use, note that it will replace any currently assigned
        sink

            >>> pipeline = Pipeline(FolderConnector(path='/tmp/', file_filter='example.pdf'))
            >>> pipeline.set_sink(ExampleSink())

        :param sink: the sink for the pipeline
        """
        logging.info(f"Setting sink {sink.get_name()} on {self.name}")
        self.sink = sink

        return self

    def run(self):
        """
        Run the current pipeline, note that you must have a sink in place to allow the pipeline to run

            >>> pipeline = Pipeline(FolderConnector(path='/tmp/', file_filter='example.pdf'))
            >>> pipeline.set_sink(ExampleSink())
            >>> pipeline.run()

        :return: The context from the run
        """
        self.context.statistics = PipelineStatistics()
        logging.info(f"Starting pipeline {self.name}")
        for document in self.connector:

            log_stream = StringIO()
            logging.basicConfig(stream=log_stream, level=self.logging_level)
            logging.info(f"Processing {document}")
            for step in self.steps:
                try:
                    if not callable(step):
                        logging.info(f"Starting step {step.get_name()}")

                        if len(signature(step.process).parameters) == 1:
                            document = step.process(document)
                        else:
                            document = step.process(document, self.context)
                    else:
                        logging.info(f"Starting step function {step.__name__}")

                        if len(signature(step).parameters) == 1:
                            document = step(document)
                        else:
                            document = step(document, self.context)
                except:
                    tt, value, tb = sys.exc_info()
                    document.exceptions.append({
                        "step": step.__name__ if callable(step) else step.get_name(),
                        "traceback": traceback.format_exception(tt, value, tb)
                    })
                    if self.stop_on_exception:
                        raise

            if self.sink:
                logging.info(f"Writing to sink {self.sink.get_name()}")
                try:
                    self.sink.sink(document)
                except:
                    document.exceptions.append({
                        "step": self.sink.get_name(),
                        "exception": sys.exc_info()[0]
                    })
                    if self.stop_on_exception:
                        raise

            document.log = log_stream.getvalue()
            self.context.statistics.processed_document(document)

        logging.info(f"Completed pipeline {self.name}")

        return self.context


class PipelineStatistics:
    """
    A set of statistics for the processed document

    documents_processed
    document_exceptions
    """

    def __init__(self):
        self.documents_processed = 0
        self.document_exceptions = 0

    def processed_document(self, document):
        """
        Update statistics based on this document completing processing

        :param document: the document that has been processed
        """
        self.documents_processed += 1

        if document.exceptions:
            self.document_exceptions += 1
